/* <!-- copyright */
/*
 * aria2 - The high speed download utility
 *
 * Copyright (C) 2015 Tatsuhiro Tsujikawa
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */
/* copyright --> */
#include "RandomStreamPieceSelector.h"
#include "BitfieldMan.h"
#include "SimpleRandomizer.h"

namespace aria2 {

RandomStreamPieceSelector::RandomStreamPieceSelector(BitfieldMan* bitfieldMan)
    : bitfieldMan_(bitfieldMan)
{
}

RandomStreamPieceSelector::~RandomStreamPieceSelector() = default;

bool RandomStreamPieceSelector::select(size_t& index, size_t minSplitSize,
                                       const unsigned char* ignoreBitfield,
                                       size_t length)
{
  size_t start = SimpleRandomizer::getInstance()->getRandomNumber(
      bitfieldMan_->countBlock());

  auto rv = bitfieldMan_->getInorderMissingUnusedIndex(
      index, start, bitfieldMan_->countBlock(), minSplitSize, ignoreBitfield,
      length);
  if (rv) {
    return true;
  }
  rv = bitfieldMan_->getInorderMissingUnusedIndex(index, 0, start, minSplitSize,
                                                  ignoreBitfield, length);
  if (rv) {
    return true;
  }
  // Fall back to inorder search because randomized search may fail
  // because of |minSplitSize| constraint.
  return bitfieldMan_->getInorderMissingUnusedIndex(index, minSplitSize,
                                                    ignoreBitfield, length);
}

void RandomStreamPieceSelector::onBitfieldInit() {}

} // namespace aria2
