/* <!-- copyright */
/*
 * aria2 - The high speed download utility
 *
 * Copyright (C) 2012 Tatsuhiro Tsujikawa
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */
/* copyright --> */
#ifndef D_WR_DISK_CACHE_H
#define D_WR_DISK_CACHE_H

#include "common.h"

#include <set>

#include "a2functional.h"

namespace aria2 {

class WrDiskCacheEntry;

class WrDiskCache {
public:
  WrDiskCache(size_t limit);
  ~WrDiskCache();
  // Adds the cache entry |ent| to the storage. The size of cached
  // data of ent is added to total_.
  bool add(WrDiskCacheEntry* ent);
  // Removes the cache entry |ent| from the storage. The size of
  // cached data of ent is subtracted from total_.
  bool remove(WrDiskCacheEntry* ent);
  // Updates the already added entry |ent|. The |delta| means how many
  // bytes is increased in this update. If the size is reduced, use
  // negative value.
  bool update(WrDiskCacheEntry* ent, ssize_t delta);
  // Evicts entries from storage so that total size of cache is kept
  // under the limit.
  void ensureLimit();
  size_t getSize() const { return total_; }

private:
  typedef std::set<WrDiskCacheEntry*, DerefLess<WrDiskCacheEntry*>> EntrySet;
  // Maximum number of bytes the storage can cache.
  size_t limit_;
  // Current number of bytes cached.
  size_t total_;
  EntrySet set_;
  int64_t clock_;
};

} // namespace aria2

#endif // D_WR_DISK_CACHE_H
